#include<stdio.h>
#include<math.h>
#include"calculation.h"

//Tridiag$B$G;H$&!#(Bhouseholder$B$K$h$C$F$G$-$?%Y%/%H%k$K$h$kD>8r9TNs$r1&$+$i3]$1$k(B
static void Mult(Matrix* x, Vector* v);

Vector *CreateVector(int size)
{
	Vector *temp;

	if((temp = (Vector*)malloc(sizeof(Vector))) == NULL){
		fprintf(stderr, "Allocation Error\n");
		exit(1);
	}

	if((temp->v = (double*)malloc(sizeof(double)*size)) == NULL){
		fprintf(stderr, "Allocation Error\n");
		exit(1);
	}

	temp->size = size;

	return temp;
}

void FreeVector(Vector* vector)
{
	free(vector->v);
	free(vector);
}

double GetDotProduct(Vector* vector1, Vector* vector2)
{
	double sum;
	int i;

	if(vector1->size != vector2->size){
		fprintf(stderr, "Cannot get dot product.\n");
		exit(1);
	}

	for(i=0, sum=.0; i<vector1->size; i++){
		sum += vector1->v[i] * vector2->v[i];
	}

	return sum;
}

Matrix *CreateMatrix(int width, int height)
{
	Matrix *temp;

	if((temp = (Matrix*)malloc(sizeof(Matrix))) == NULL){
		fprintf(stderr, "Allocation Error\n");
		exit(1);
	}

	if((temp->a = (double*)malloc(sizeof(double)*width*height)) == NULL){
		fprintf(stderr, "Allocation Error\n");
		exit(1);
	}

	temp->width = width;
	temp->height = height;

	return temp;
}

void FreeMatrix(Matrix* matrix)
{
	free(matrix->a);
	free(matrix);
}

double Householder(Vector *x)
{
	double norm;
	double weight;
	int i, j;
	

	norm = sqrt(GetDotProduct(x, x));

	if(norm){
		if(x->v[0] < 0) norm = - norm;
		x->v[0] += norm;

		weight = 1/(sqrt(2*norm*x->v[0]));

		for(i=0; i<x->size; i++){
			x->v[i] *= weight;
		}
	}

	return -norm;
}


void Tridiag(Matrix *m)
{
	Vector *v;
	int i, j;
	double tmp;

	v = CreateVector(m->width);

	for(i=0; i<m->width-2; i++){
		v->size--;
		for(j=i+1; j<m->width; j++){
			v->v[j-i-1] = m->a[i*m->width + j];
		}
	  if(!(tmp = Householder(v))) continue;

		Mult(m, v);

		m->a[i*m->width + i+1] = m->a[(i+1)*m->width + i] = tmp;
		for(j=i+2; j<m->width; j++){
			m->a[i*m->width + j] = m->a[j*m->width + i] = 0;
		}
		

	}

	FreeVector(v);
}

static void Mult(Matrix* x, Vector* v)
{
	int i, j, offset;
	double tmp;
	Vector *g;
  
	offset = x->width - v->size;
	g = CreateVector(v->size);

	for(i=offset; i<x->width; i++){
		g->v[i-offset] = 0;
		for(j=offset; j<x->width; j++){
			g->v[i-offset] += x->a[i*x->width + j]*v->v[j-offset];
		}
	}

	tmp = GetDotProduct(g, v);

	for(j=0; j<g->size; j++){
		g->v[j] = 2*(g->v[j] - v->v[j]*tmp);
	}
	for(i=offset; i<x->width; i++){
		for(j=offset; j<x->width; j++){
			x->a[i*x->width +j] -= (v->v[i-offset]*g->v[j-offset] + g->v[i-offset]*v->v[j-offset]);
		}
	}

	FreeVector(g);
}
